//
//  GameScene.swift
//  Flappy Felipe
//
//  Created by Main Account on 2/20/15.
//  Copyright (c) 2015 Razeware LLC. All rights reserved.
//

import SpriteKit

enum Layer: CGFloat {
  case Background
  case Obstacle
  case Foreground
  case Player
}

class GameScene: SKScene {

  let kGravity: CGFloat = -1500.0
  let kImpulse: CGFloat = 400.0
  let kNumForegrounds = 2
  let kGroundSpeed: CGFloat = 150.0
  let kBottomObstacleMinFraction: CGFloat = 0.1
  let kBottomObstacleMaxFraction: CGFloat = 0.6
  let kGapMultiplier: CGFloat = 3.5
  let kFirstSpawnDelay: NSTimeInterval = 1.75
  let kEverySpawnDelay: NSTimeInterval = 1.5
  
  let worldNode = SKNode()
  var playableStart: CGFloat = 0
  var playableHeight: CGFloat = 0
  let player = SKSpriteNode(imageNamed: "Bird0")
  var lastUpdateTime: NSTimeInterval = 0
  var dt: NSTimeInterval = 0
  var playerVelocity = CGPoint.zeroPoint
  let sombrero = SKSpriteNode(imageNamed: "Sombrero")
  
  let dingAction = SKAction.playSoundFileNamed("ding.wav", waitForCompletion: false)
  let flapAction = SKAction.playSoundFileNamed("flapping.wav", waitForCompletion: false)
  let whackAction = SKAction.playSoundFileNamed("whack.wav", waitForCompletion: false)
  let fallingAction = SKAction.playSoundFileNamed("falling.wav", waitForCompletion: false)
  let hitGroundAction = SKAction.playSoundFileNamed("hitGround.wav", waitForCompletion: false)
  let popAction = SKAction.playSoundFileNamed("pop.wav", waitForCompletion: false)
  let coinAction = SKAction.playSoundFileNamed("coin.wav", waitForCompletion: false)

  override func didMoveToView(view: SKView) {
    addChild(worldNode)
    setupBackground()
    setupForeground()
    setupPlayer()
    setupSombrero()
    startSpawning()
  }
  
  // MARK: Setup methods
  
  func setupBackground() {
    
    let background = SKSpriteNode(imageNamed: "Background")
    background.anchorPoint = CGPoint(x: 0.5, y: 1.0)
    background.position = CGPoint(x: size.width/2, y: size.height)
    background.zPosition = Layer.Background.rawValue
    worldNode.addChild(background)
    
    playableStart = size.height - background.size.height
    playableHeight = background.size.height
  
  }
  
  func setupForeground() {
  
    for i in 0..<kNumForegrounds {
      let foreground = SKSpriteNode(imageNamed: "Ground")
      foreground.anchorPoint = CGPoint(x: 0, y: 1)
      foreground.position = CGPoint(x: CGFloat(i) * size.width, y: playableStart)
      foreground.zPosition = Layer.Foreground.rawValue
      foreground.name = "foreground"
      worldNode.addChild(foreground)
    }
  
  }
  
  func setupPlayer() {
  
    player.position = CGPoint(x: size.width * 0.2, y: playableHeight * 0.4 + playableStart)
    player.zPosition = Layer.Player.rawValue
    worldNode.addChild(player)
  
  }
  
  func setupSombrero() {
   
    sombrero.position = CGPoint(x: 31 - sombrero.size.width/2, y: 29 - sombrero.size.height/2)
    player.addChild(sombrero)
    
  }
  
  // MARK: Gameplay
  
  func createObstacle() -> SKSpriteNode {
    let sprite = SKSpriteNode(imageNamed: "Cactus")
    sprite.zPosition = Layer.Obstacle.rawValue
    return sprite
  }
  
  func spawnObstacle() {
  
    let bottomObstacle = createObstacle()
    let startX = size.width + bottomObstacle.size.width/2
    
    let bottomObstacleMin = (playableStart - bottomObstacle.size.height/2) + playableHeight * kBottomObstacleMinFraction
    let bottomObstacleMax = (playableStart - bottomObstacle.size.height/2) + playableHeight * kBottomObstacleMaxFraction
    bottomObstacle.position = CGPointMake(startX, CGFloat.random(min: bottomObstacleMin, max: bottomObstacleMax))
    worldNode.addChild(bottomObstacle)
    
    let topObstacle = createObstacle()
    topObstacle.zRotation = CGFloat(180).degreesToRadians()
    topObstacle.position = CGPoint(x: startX, y: bottomObstacle.position.y + bottomObstacle.size.height/2 + topObstacle.size.height/2 + player.size.height * kGapMultiplier)
    worldNode.addChild(topObstacle)
    
    let moveX = size.width + topObstacle.size.width
    let moveDuration = moveX / kGroundSpeed
    let sequence = SKAction.sequence([
      SKAction.moveByX(-moveX, y: 0, duration: NSTimeInterval(moveDuration)),
      SKAction.removeFromParent()
    ])
    topObstacle.runAction(sequence)
    bottomObstacle.runAction(sequence)
  
  }
  
  func startSpawning() {
  
    let firstDelay = SKAction.waitForDuration(kFirstSpawnDelay)
    let spawn = SKAction.runBlock(spawnObstacle)
    let everyDelay = SKAction.waitForDuration(kEverySpawnDelay)
    let spawnSequence = SKAction.sequence([
      spawn, everyDelay
    ])
    let foreverSpawn = SKAction.repeatActionForever(spawnSequence)
    let overallSequence = SKAction.sequence([firstDelay, foreverSpawn])
    runAction(overallSequence)
  
  }
  
  func flapPlayer() {
  
    // Play sound
    runAction(flapAction)

    // Apply impulse
    playerVelocity = CGPoint(x: 0, y: kImpulse)
    
    // Move sombrero
    let moveUp = SKAction.moveByX(0, y: 12, duration: 0.15)
    moveUp.timingMode = .EaseInEaseOut
    let moveDown = moveUp.reversedAction()
    sombrero.runAction(SKAction.sequence([moveUp, moveDown]))
    
  }
  
  override func touchesBegan(touches: NSSet, withEvent event: UIEvent) {
    flapPlayer()
  }
  
  // MARK: Updates
  
  override func update(currentTime: CFTimeInterval) {
    if lastUpdateTime > 0 {
      dt = currentTime - lastUpdateTime
    } else {
      dt = 0
    }
    lastUpdateTime = currentTime
    
    updatePlayer()
    updateForeground()
  }
  
  func updatePlayer() {
  
    // Apply gravity
    let gravity = CGPoint(x: 0, y: kGravity)
    let gravityStep = gravity * CGFloat(dt)
    playerVelocity += gravityStep
    
    // Apply velocity
    let velocityStep = playerVelocity * CGFloat(dt)
    player.position += velocityStep
    
    // Temporary halt when hits ground
    if player.position.y - player.size.height/2 < playableStart {
      player.position = CGPoint(x: player.position.x, y: playableStart + player.size.height/2)
    }
  
  }
  
  func updateForeground() {
  
    worldNode.enumerateChildNodesWithName("foreground", usingBlock: { node, stop in
      if let foreground = node as? SKSpriteNode {
        let moveAmt = CGPoint(x: -self.kGroundSpeed * CGFloat(self.dt), y: 0)
        foreground.position += moveAmt
        
        if foreground.position.x < -foreground.size.width {
          foreground.position += CGPoint(x: foreground.size.width * CGFloat(self.kNumForegrounds), y: 0)
        }
        
      }
    })
  
  }
  
}
